@@ -102,7 +102,7 @@ static GHashTable    *global_fh_table     = NULL;
 static void
 debug_print (const gchar *message, ...)
 {
-#ifdef DEBUG_ENABLED
+#if DEBUG_ENABLED
 
   static FILE *debug_fhd = NULL;
   va_list      var_args;
@@ -1315,7 +1315,7 @@ vfs_flush (const gchar *path, struct fuse_file_info *fi)
   debug_print ("vfs_flush: %s\n", path);
 
   return 0;
-};
+}
 
 static gint
 vfs_opendir (const gchar *path, struct fuse_file_info *fi)
